#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "../include/tdsio.h"
#include "../include/tdssdh.h"
#include "examples.h"

#define TIME_CHECK    

#define MAX_FILE_SIZE		(1 * 1024 * 1024 * 1024 /* 1Gbyte */)
#define DEFAULT_BLOCK_SIZE	(64 * 1024 /* 64kbyte */)
#define DEFAULT_DEVFILE		"/dev/tds0"
#define DEFAULT_OUTFILE		"./tds.data"


static const char *progName;

static void usage(void);
static void
usage()
{
	fprintf(stderr, "usage: %s [-vI] "
	    "[-C channes] [-F filter] [-R rate] "
#ifdef  TIME_CHECK
	    "[-T sync] "
#endif
	    "[-W bit-width] [-a auxfile] [-b block] "
	    "[-d dcoff] [-i device] [-n repeat] [-o file] [-t type]\n",
	    progName);
	fprintf(stderr, "       type : is defined 9260, 9270\n");
	exit(EXIT_FAILURE);
}

int main(int argc, char **argv)
{
	mode_t ofilemode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
	struct tdsauxio sdhio;
	tdsdev_t *tdsdev = NULL;
	tds_time_t tdstime;
	uint32_t year, status;
	uint32_t samplingRate, samplingFilter, samplingWidth, samplingChannels;
	uint32_t drv_version, lib_version;
	unsigned int dcoff[4], dcoff2[4];
	int samplingVerbose, samplingCount, samplingBlock, syncpattern;
	int fd_out, opt, fileno, rv, dc, i, initflag;
#ifdef	TIME_CHECK
	int checkTimeSync = 0;
#endif
	char *devfile, *outfile, *auxfile, *type, *buf;
	extern int optind;
	extern char *optarg;

	samplingRate = TDSIO_SAMPLING_16MHZ;
	samplingFilter = TDSIO_SAMPLING_THRU;
	samplingWidth = TDSIO_SAMPLING_8BIT;
	samplingChannels = TDSIO_SAMPLING_1CH;
	samplingVerbose = 0;
	samplingCount = -1;
	samplingBlock = DEFAULT_BLOCK_SIZE;
	fileno = 0;
	dc = 0;
	devfile = DEFAULT_DEVFILE;
	outfile = DEFAULT_OUTFILE;
	auxfile = NULL;
	type = "";
	initflag = 0;

	/* ޥɹԥץβ */
	progName = argv[0];
	while((opt = getopt(argc, argv, "C:F:IR:T:W:a:b:d:i:n:o:t:v")) != EOF) {
		switch(opt) {
		case 'C':	/* number of channel */
			samplingChannels = get_channels(optarg);
			break;
		case 'F':	/* number of channel */
			samplingFilter = get_filter(optarg);
			break;
		case 'I':	/* must initialize */
			initflag = 1;
			break;
		case 'R':	/* sampling rate */
			samplingRate = get_rate(optarg);
			break;
#ifdef  TIME_CHECK
		case 'T':
			checkTimeSync = 1;
			syncpattern = strtoul(optarg, NULL, 0);
			break;
#endif
		case 'W':	/* sampling bit width */
			samplingWidth = get_width(optarg);
			break;
		case 'a':	/* read block size */
			auxfile = optarg;
			break;
		case 'b':	/* read block size */
			samplingBlock = (int)strtoul(optarg, NULL, 0);
			break;
		case 'd':	/* DCoffset */
			dcoff[dc++] = (int)strtoul(optarg, NULL, 0);
			break;
		case 'i':	/* input device */
			devfile = optarg;
			break;
		case 'n':	/* repeat count */
			samplingCount = (int)strtol(optarg, NULL, 0);
			break;
		case 'o':	/* output file */
			outfile = optarg;
			break;
		case 't':	/* board type */
			for (i = 0; typetbl[i] != NULL; i ++)
				if (strcmp(optarg, typetbl[i]) == 0)
					break;
			if (typetbl[i] == NULL)
				usage();
			type = typetbl[i];
			break;
		case 'v':	/* verbose mode */
			samplingVerbose = 1;
			break;
		case '?':
		default:
			usage();
		}
	}

	/* ǥХΥץ */
	if ((tdsdev = TdsOpen(devfile)) == NULL) {
		perror("TdsOpen() failed ");
		return -1;
	} else
		printf("TdsOpen() OK\n");
	if ((rv = TdsGetVersion(tdsdev, &drv_version, &lib_version)) != 0) {
		perror("TdsGetVersion() failed ");
		goto end;
	} else {
		printf("%s: driver version: %d.%d.%d\n", devfile,
		    TDS_VERSION_MAJOR(drv_version),
		    TDS_VERSION_MINOR(drv_version),
		    TDS_VERSION_REVISION(drv_version));
		printf("%s: library version: %d.%d.%d\n", devfile,
		    TDS_VERSION_MAJOR(lib_version),
		    TDS_VERSION_MINOR(lib_version),
		    TDS_VERSION_REVISION(lib_version));
	}

	/* ϥեΥץ */
	if (strcmp(outfile, "-") == 0)
		fd_out = 1;
	else if (outfile != NULL) {
		if ((fd_out =
		    open(outfile, O_WRONLY | O_CREAT | O_TRUNC, ofilemode)) ==
		    -1) {
			perror("output file open() error");
			goto end;
		}
	}

	/* ǥХ */
	if ((rv = TdsInit(tdsdev)) != 0) {
		perror("TdsInit() failed ");
		return -1;
	} else
		printf("TdsInit() OK\n");

	/* ץ */
	if ((rv = TdsSetSampling2(tdsdev, &samplingRate,
	    &samplingFilter, &samplingWidth, &samplingChannels)) != 0) {
		perror("TdsSetSampling2() failed ");
		goto end;
	} else
		printf("TdsSetSampling2() OK\n");

	/* ץ󥰥⡼ɼ */
	if ((rv = TdsGetSampling2(tdsdev, &samplingRate,
	    &samplingFilter, &samplingWidth, &samplingChannels)) != 0) {
		perror("TdsGetSampling() failed ");
		goto end;
	} else {
		printf("TdsGetSampling() OK\n");
		printf("%dHz/%dbit/%dch\n",
		    value2key(samplingRate, Freq),
		    value2key(samplingWidth, Bitw),
		    value2key(samplingChannels, Chan));
		if (samplingFilter != -1)
			printf("filter %d\n",
			    value2key(samplingFilter, Filter));
	}

	if (auxfile != NULL) {
		if ((rv = TdsSetAuxFile(tdsdev, auxfile)) != 0) {
			perror("TdsSetAuxFile() failed ");
			return -1;
		} else
			printf("TdsSetAuxFile() OK\n");
	}
	if (strcmp(type, "9270") == 0) {
		if ((rv = TdsGetAux(tdsdev, &sdhio.auxsize, sdhio.auxfield)) !=
		    0) {
			perror("TdsGetAux() failed ");
		} else
			printf("TdsGetAux() OK\n");
	}

	/* ǥХFIFOꥢ */
	if ((rv = TdsClrBuffer(tdsdev)) != 0) {
		perror("TdsClrBuffer() failed ");
		goto end;
	} else
		printf("TdsClrBuffer() OK\n");

	if (initflag) {
		/* 1PPSƱ */
		if ((rv = TdsSync1pps(tdsdev)) != 0) {
			perror("TdsSync1pps() failed ");
			goto end;
		} else
			printf("TdsSync1pps() OK\n");

		/* Ʊ */
		if (strcmp(type, "9270") == 0 ||
	    	(rv = TdsSyncTime(tdsdev)) != 0) {
			if (rv != 0 && errno == ENOTTY)
				perror("ioctl() time sync not support ");
			else if (rv != 0)
				perror("ioctl() time sync failed ");

			if ((rv = TdsSetTime(tdsdev, NULL)) != 0) {
				perror("TdsSetTime() failed ");
				goto end;
			} else
				printf("TdsSetTime() OK\n");

			if ((rv = TdsSetYear(tdsdev, NULL)) != 0) {
				perror("TdsSetYear() failed ");
				goto end;
			} else
				printf("TdsSetYear() OK\n");
		} else
			printf("TdsSyncTime() OK\n");
	}

	/*  */
	if ((rv = TdsGetTime(tdsdev, &tdstime)) != 0) {
		perror("TdsGetTime() failed ");
		goto end;
	} else
		printf("TdsGetTime() OK\n");

	if ((rv = TdsGetYear(tdsdev, &year)) != 0) {
		int err;
		char *msg;
		perror("TdsGetYear() failed ");
		rv = TdsError(tdsdev, &err, &msg);
		if (rv != 0) {
			perror("TdsError() failed ");
		} else {
			printf("TdsError() OK\n");
			printf("ERROR: (%d) %s\n", err, msg);
		}
		goto end;
	}

	if (dc > 0) {
		if ((rv = TdsSetDcoffset(tdsdev, &dc, dcoff)) != 0) {
			perror("TdsSetDcoffset() failed ");
			goto end;
		}
		printf("TdsSetDcoffset() OK\n");
		if ((rv = TdsGetDcoffset(tdsdev, &dc, dcoff2)) != 0) {
			perror("TdsGetDcoffset() failed ");
			goto end;
		}
		printf("TdsGetDcoffset() OK\n");
		for (i = 0; i < 4; i++)
			if (dcoff[i] != dcoff2[i]) {
				perror("DCoffset no match ");
				goto end;
			}
		printf("DCoffset match\n");
	}

	/* ץ󥰳 */
	if ((rv = TdsStart(tdsdev)) != 0) {
		perror("TdsStart() failed ");
		goto end;
	} else
		printf("TdsStart() OK\n");

	if ((rv = TdsStatus(tdsdev, &status)) != 0) {
		perror("TdsStatus() failed\n");
		goto end;
	} else
		printf("TdsStatus() OK\n");

	buf = (char *)malloc(samplingBlock);
	if (buf == NULL) {
		perror("malloc() failed ");
		goto end;
	}
	/* ǡɤ߹ */
	{
#ifdef	TIME_CHECK
#define ROUNDUP(val,bound)	((((val) + (bound) - 1) / (bound)) * (bound))
		tdssdh_t h;
		size_t total_bytes = 0;
		size_t bytespersec = 0;
		size_t sync_pos;
		u_int32_t last_time;
#endif	/*TIME_CHECK*/
		size_t fsize = 0;
		register int i;
		for(i = 0; i != samplingCount; i++) {
			size_t nbyte = samplingBlock;

			if ((rv = TdsGetData(tdsdev, (void *)buf)) != 0) {
				perror("TdsGetData() failed ");
				goto end;
			} else
				if (samplingVerbose)
					printf("TdsGetData() OK\n");
			nbyte -= sizeof(uint32_t);
			if ((rv = TdsRead(tdsdev, &nbyte,
			    (unsigned char *)(buf + sizeof (uint32_t)))) != 0) {
				perror("TdsRead() failed ");
				goto end;
			}
			nbyte += sizeof(uint32_t);
			if (samplingVerbose) {
				printf(".");
				if ((i % 50) == 49) printf(": %d\n", i+1);
				fflush(stdout);
			}
			/* ǡν񤭽Ф */
			if (strcmp(outfile, "-") == 0 ||
			    fsize + nbyte < MAX_FILE_SIZE)
				rv = write(fd_out, buf, nbyte);
			else {
				char path[MAXPATHLEN];

				rv = write(fd_out, buf, MAX_FILE_SIZE - fsize);
				if (rv != MAX_FILE_SIZE - fsize) {
					perror("write() failed ");
					goto end;
				}
				close(fd_out);
				nbyte -= rv;
				fsize = 0;
				fileno++;

				snprintf(path,
				    sizeof (path), "%s%d", outfile, fileno);
				fd_out = open(path,
				    O_WRONLY | O_CREAT | O_TRUNC, ofilemode);
				if (fd_out == -1) {
					perror("output file open() error");
					goto end;
				}
				rv = write(fd_out,
				    buf + rv / sizeof (u_int32_t), nbyte);
			}
			if (rv != nbyte) {
				perror("write() failed ");
				goto end;
			}
			fsize += nbyte;
#ifdef	TIME_CHECK
			if (!checkTimeSync)
			    continue;
			/* ƱǡΥå */
			if (bytespersec == 0) {
				bytespersec =
				    sizeof (struct tdssdh) +
				    sizeof (uint16_t) + sdhio.auxsize +
				    value2key(samplingRate, Freq) *
				    value2key(samplingWidth, Bitw) *
				    value2key(samplingChannels, Chan) / 8;
				h = (tdssdh_t)&buf;
				last_time =
				    (h->totalsec0 || h->totalsec1 << 16) - 1;
			}
			sync_pos = ROUNDUP(total_bytes, bytespersec);
			while(sync_pos < nbyte) {
				u_int32_t curr_time;

				h = (tdssdh_t)&buf[sync_pos];
				if (h->syncptn0 == 0xffff &&
				    h->syncptn1 == 0xffff &&
				    h->syncptn2 == syncpattern) {
					printf(
					    "%d: sync pattern not found\n", i);
					goto end;
				}
				curr_time = h->totalsec0 | h->totalsec1 << 16;
				if (curr_time != last_time + 1) {
					printf(
					    "%d: uncontig time-of-day value\n",
					    i);
					goto end;
				}
				last_time = curr_time;
				sync_pos += bytespersec;
			}
			total_bytes += nbyte;
			total_bytes %= bytespersec;
#endif	/*TIME_CHECK*/
		}
	}

	/* ץ */
	rv = TdsStop(tdsdev);
	if (rv != 0) {
		perror("TdsStop() failed ");
		goto end;
	}
	else {
		printf("TdsStop() OK\n");
	}

 end:
	/* ϥեΥ */
	close(fd_out);
	/* ǥХΥ */
	TdsClose(tdsdev);

	return rv;
}
